import os
import logging
import pandas as pd
from datetime import datetime
from dotenv import load_dotenv
from selenium import webdriver
from selenium.webdriver.chrome.service import Service
from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.chrome.options import Options
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from concurrent.futures import ThreadPoolExecutor, as_completed
import time

# Load environment variables from .env file located in the config directory
config_dir = 'C:\\Users\\user\\projects\\st\\config'
load_dotenv(dotenv_path=os.path.join(config_dir, '.env'))

settings = {
    'data_path': os.getenv('DATA_PATH')
}

logging.basicConfig(level=logging.INFO)

def get_stock_codes_and_names():
    file_path = os.path.join(settings['data_path'], 'stock_codes.csv')
    if not os.path.exists(file_path):
        logging.error(f"{file_path} not found.")
        return []
    
    df = pd.read_csv(file_path)
    return df.to_dict('records')

def parse_chosun_news(driver, stock_name):
    news_items = []
    try:
        driver.get('https://www.chosun.com/nsearch/?query=')
        time.sleep(2)  # 페이지 로딩을 위한 대기 시간

        search_box = driver.find_element(By.CSS_SELECTOR, 'input.search-bar--input')
        search_box.send_keys(stock_name)
        search_box.send_keys(Keys.RETURN)
        time.sleep(2)  # 검색 결과 로딩을 위한 대기 시간

        articles = []

        article_selectors = [
            'div.search_result dl dt a',  # 일반적인 기사 제목
            'div.story-card h2 a',        # 스토리 카드 형식의 기사 제목
            'div.story-card a'            # 다른 형식의 기사 제목
        ]

        for selector in article_selectors:
            articles.extend(driver.find_elements(By.CSS_SELECTOR, selector))

        for article in articles:
            title = article.text.strip()  # 제목에서 불필요한 공백 제거
            link = article.get_attribute('href')
            if title and "기자" not in title:  # 제목이 비어 있지 않고 '기자'라는 단어가 포함되지 않은 경우만 출력
                news_items.append({'title': title, 'link': link})
    except Exception as e:
        logging.error(f"Error occurred while extracting news from Chosun: {e}")
    return news_items

def parse_yonhap_news(driver, stock_name):
    news_items = []
    retries = 3  # 재시도 횟수
    for attempt in range(retries):
        try:
            driver.get("https://www.yonhapnewstv.co.kr")

            wait = WebDriverWait(driver, 10)
            search_button = wait.until(EC.element_to_be_clickable((By.CLASS_NAME, "btnOpenSearch")))
            driver.execute_script("arguments[0].click();", search_button)

            search_bar = wait.until(EC.presence_of_element_located((By.ID, "inputSearch")))
            driver.execute_script("arguments[0].value=arguments[1];", search_bar, stock_name)
            search_bar.send_keys(Keys.RETURN)

            time.sleep(5)

            news_elements = driver.find_elements(By.CSS_SELECTOR, 'ul.article.article-webzine > li.item')

            for element in news_elements:
                title_element = element.find_element(By.CSS_SELECTOR, 'a.title')
                link_element = element.find_element(By.CSS_SELECTOR, 'a.item-img')
                title = title_element.text
                link = "https://www.yonhapnewstv.co.kr" + link_element.get_attribute('href')
                news_items.append({'title': title, 'link': link})
            return news_items
        except Exception as e:
            logging.error(f"Error occurred while extracting news from Yonhap on attempt {attempt + 1}: {e}")
            if attempt < retries - 1:
                time.sleep(5)  # 재시도 전에 잠시 대기
            else:
                logging.error(f"Failed to fetch news from Yonhap after {retries} attempts.")
    return news_items

def update_news_data(stock_code, stock_name):
    driver_path = r'C:\tools\chromedriver\chromedriver.exe'
    options = webdriver.ChromeOptions()
    options.add_argument('--headless')
    options.add_argument('--disable-gpu')
    service = Service(executable_path=driver_path)

    all_news_data = []
    
    # 조선일보 뉴스 수집
    driver = webdriver.Chrome(service=service, options=options)
    try:
        logging.info(f"Fetching news from Chosun for stock name: {stock_name}")
        news_items = parse_chosun_news(driver, stock_name)
        for item in news_items:
            news_data = {
                'title': item['title'],
                'link': item['link'],
                'source': 'Chosun',
                'date': datetime.now().strftime('%Y-%m-%d'),
                'stock_code': stock_code,
                'stock_name': stock_name
            }
            all_news_data.append(news_data)
        logging.info(f"Fetched {len(news_items)} news items from Chosun for stock name: {stock_name}")
    finally:
        driver.quit()
    
    # 연합뉴스 뉴스 수집
    driver = webdriver.Chrome(service=service, options=options)
    try:
        logging.info(f"Fetching news from Yonhap for stock name: {stock_name}")
        news_items = parse_yonhap_news(driver, stock_name)
        for item in news_items:
            news_data = {
                'title': item['title'],
                'link': item['link'],
                'source': 'Yonhap',
                'date': datetime.now().strftime('%Y-%m-%d'),
                'stock_code': stock_code,
                'stock_name': stock_name
            }
            all_news_data.append(news_data)
        logging.info(f"Fetched {len(news_items)} news items from Yonhap for stock name: {stock_name}")
    finally:
        driver.quit()

    if not all_news_data:
        logging.warning(f"No news data fetched for {stock_name}.")
        return

    file_path = os.path.join(settings['data_path'], f"news_data_{stock_code}.csv")
    if not os.path.exists(file_path):
        logging.warning(f"File does not exist: {file_path}. Creating a new one.")
        existing_news_data = pd.DataFrame()
    else:
        try:
            existing_news_data = pd.read_csv(file_path)
            if existing_news_data.empty:
                logging.warning(f"File is empty: {file_path}. Creating a new one.")
                existing_news_data = pd.DataFrame()
        except pd.errors.EmptyDataError:
            logging.warning(f"File is empty: {file_path}. Creating a new one.")
            existing_news_data = pd.DataFrame()
        except Exception as e:
            logging.error(f"Error reading file {file_path}: {e}")
            return

    df = pd.DataFrame(all_news_data)
    combined_data = pd.concat([existing_news_data, df], ignore_index=True)

    combined_data.to_csv(file_path, index=False)
    logging.info(f"News data for {stock_name} saved to {file_path}")

def main():
    stock_codes_and_names = get_stock_codes_and_names()
    if not stock_codes_and_names:
        logging.error("No stock codes and names found.")
        return

    with ThreadPoolExecutor(max_workers=32) as executor:  # 32개의 작업자 스레드 사용
        futures = [executor.submit(update_news_data, stock['code'], stock['name']) for stock in stock_codes_and_names]
        for future in futures:
            future.result()

if __name__ == "__main__":
    main()
