import os
import json
import time
import requests
import numpy as np
import pandas as pd
import websocket
import threading
from datetime import datetime, timedelta
from sklearn.preprocessing import StandardScaler
from tensorflow.keras.models import load_model
import pickle
import logging
import sys
from Crypto.Cipher import AES
import base64
# 현재 스크립트의 경로를 기준으로 settings.py의 경로를 추가합니다.
current_dir = os.path.dirname(os.path.abspath(__file__))
project_root = os.path.abspath(os.path.join(current_dir, '..', '..'))
config_dir = os.path.join(project_root, 'config')

if config_dir not in sys.path:
    sys.path.insert(0, config_dir)
from settings import settings  # Import the settings from settings.py

# Configure logging
logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s - %(message)s')

# Initialize paths
data_dir = settings.get('data_path', os.path.join(os.path.dirname(__file__), '..', 'data'))
stock_model_path = os.path.join(data_dir, 'models', 'lstm_model_stock.h5')
stock_scaler_path = os.path.join(data_dir, 'models', 'scaler_stock.pkl')
all_model_path = os.path.join(data_dir, 'models', 'lstm_model_all.h5')
all_scaler_path = os.path.join(data_dir, 'models', 'scaler_all.pkl')

# Check if the necessary settings are loaded
required_keys = ['appkey', 'appsecret', 'token_path', 'trading_url']
missing_keys = [key for key in required_keys if not settings.get(key)]

if missing_keys:
    raise ValueError(f"Missing required settings: {', '.join(missing_keys)}")

# 1. 모델 및 스케일러 로드
def load_lstm_model_and_scaler(model_path, scaler_path):
    try:
        model = load_model(model_path)
        with open(scaler_path, 'rb') as f:
            scaler = pickle.load(f)
        return model, scaler
    except Exception as e:
        logging.error(f"Failed to load model or scaler: {e}")
        raise

# 2. 실시간 데이터 수집
def get_stock_codes_from_csv(file_path):
    """CSV 파일에서 종목 코드를 읽어옵니다."""
    try:
        if not os.path.exists(file_path):
            raise FileNotFoundError(f"CSV 파일을 찾을 수 없습니다: {file_path}")
        
        df = pd.read_csv(file_path, dtype={'code': str}, usecols=['code'])
        
        if df.empty:
            raise ValueError("CSV 파일이 비어 있습니다.")
        
        return df['code'].tolist()
        
    except Exception as e:
        logging.error(f"CSV 파일에서 종목 코드를 읽는 중 오류가 발생했습니다: {e}")
        return []

def get_valid_token():
    """유효한 토큰을 가져옵니다. 없거나 만료된 경우 새로 갱신합니다."""
    token_path = settings['token_path']
    try:
        with open(token_path, 'r') as file:
            token_data = json.load(file)
            if is_token_valid(token_data):
                return token_data['access_token']
            else:
                return refresh_token()
    except (FileNotFoundError, json.JSONDecodeError):
        return refresh_token()

def is_token_valid(token_data):
    """토큰이 여전히 유효한지 확인합니다."""
    expiration = datetime.strptime(token_data['expiration'], '%Y-%m-%d %H:%M:%S')
    return expiration > datetime.now()

def refresh_token():
    """만료되었거나 유효하지 않은 경우 토큰을 갱신합니다."""
    url = f"{settings['trading_url']}/oauth2/tokenP"
    data = {
        'grant_type': 'client_credentials',
        'appkey': settings['appkey'],
        'appsecret': settings['appsecret']
    }
    headers = {'Content-Type': 'application/json'}
    try:
        response = requests.post(url, headers=headers, json=data)
        if response.status_code == 200:
            response_data = response.json()
            token_data = {
                'access_token': response_data['access_token'],
                'expiration': (datetime.now() + timedelta(seconds=response_data['expires_in'])).strftime('%Y-%m-%d %H:%M:%S')
            }
            with open(settings['token_path'], 'w') as file:
                json.dump(token_data, file)
            return token_data['access_token']
        else:
            raise Exception(f"토큰 갱신 실패: {response.text}")
    except requests.exceptions.RequestException as e:
        raise Exception(f"토큰 갱신 실패: {e}")

def decrypt_data(encrypted_data, key, iv):
    """AES256을 사용하여 데이터를 복호화합니다."""
    try:
        cipher = AES.new(key.encode('utf-8'), AES.MODE_CBC, iv.encode('utf-8'))
        decrypted = cipher.decrypt(base64.b64decode(encrypted_data))
        return decrypted.decode('utf-8').rstrip("\0")
    except Exception as e:
        logging.error(f"Failed to decrypt data: {e}")
        return None

def on_message(ws, message):
    """실시간 WebSocket 메시지를 처리합니다."""
    try:
        data = json.loads(message)
        if data.get('body', {}).get('msg1') == "SUBSCRIBE SUCCESS":
            logging.info("성공적으로 구독되었습니다.")
        elif data.get('body', {}).get('output', {}).get('iv'):
            iv = data['body']['output']['iv']
            key = data['body']['output']['key']
            encrypted_data = data['body']['output']['data']
            decrypted_data = decrypt_data(encrypted_data, key, iv)
            if decrypted_data:
                process_real_time_data(json.loads(decrypted_data))
        else:
            process_real_time_data(data)
    except json.JSONDecodeError:
        logging.error("WebSocket 메시지 디코딩 중 오류가 발생했습니다.")
    except KeyError as e:
        logging.error(f"필수 데이터가 누락되었습니다: {e}")

def on_error(ws, error):
    """WebSocket 오류를 처리합니다."""
    logging.error(f"WebSocket 오류 발생: {error}")

def on_close(ws, close_status_code, close_msg):
    """WebSocket 연결 종료를 처리합니다."""
    logging.warning(f"WebSocket이 종료되었습니다: {close_status_code} - {close_msg}")
    for attempt in range(3):
        logging.info(f"Reconnecting... Attempt {attempt + 1}")
        try:
            time.sleep(5 * (2 ** attempt))  # Exponential backoff
            run_websocket()
            break
        except Exception as e:
            logging.error(f"Reconnection attempt {attempt + 1} failed: {e}")
            if attempt == 2:
                logging.error("Max reconnection attempts reached. Exiting.")
                break

def on_open(ws):
    """WebSocket 연결이 열릴 때 구독 메시지를 전송합니다."""
    def run():
        approval_key = get_valid_token()
        stock_codes = get_stock_codes_from_csv(os.path.join(data_dir, 'stock_codes.csv'))

        for stock_code in stock_codes:
            subscribe_message = {
                "header": {
                    "approval_key": approval_key,
                    "custtype": "P",
                    "tr_type": "1",
                    "content-type": "utf-8"
                },
                "body": {
                    "input": {
                        "tr_id": "H0STASP0",
                        "tr_key": stock_code
                    }
                }
            }
            ws.send(json.dumps(subscribe_message))
            time.sleep(1)  # 각 요청 간에 딜레이 추가
    thread = threading.Thread(target=run)
    thread.start()

# 3. 데이터 수집 대기 및 예측 수행
def create_dataset(data, time_step):
    """LSTM 모델을 위한 데이터셋을 생성합니다."""
    X, y = [], []
    for i in range(len(data) - time_step):
        X.append(data[i:(i + time_step), 0])
        y.append(data[i + time_step, 0])
    return np.array(X), np.array(y)

def update_real_time_model(model, scaler, ticker, time_step=60):
    """새로운 데이터로 실시간 모델을 업데이트합니다."""
    real_time_data = get_real_time_data_with_retry(ticker)
    if real_time_data is None:
        return model
    
    if 'close' not in real_time_data.columns:
        real_time_data = real_time_data[['stck_prpr']].rename(columns={'stck_prpr': 'close'})
    
    if len(real_time_data) < time_step:
        logging.warning(f"주어진 time_step={time_step}에 대한 데이터가 충분하지 않습니다. 데이터 길이={len(real_time_data)}")
        return model
    
    real_time_data, _ = normalize_data(real_time_data)
    scaled_real_time_data = real_time_data['close'].values.reshape(-1, 1)
    
    X_real_time, y_real_time = create_dataset(scaled_real_time_data, time_step)
    if len(X_real_time) == 0:
        logging.warning("데이터셋 생성 후 충분한 데이터 포인트가 없습니다.")
        return model
    
    X_real_time = X_real_time.reshape(X_real_time.shape[0], X_real_time.shape[1], 1)
    
    model = update_lstm_model(pd.DataFrame(y_real_time), model, time_step)
    return model

def process_real_time_data(data):
    """실시간 데이터를 처리하여 예측을 수행합니다."""
    try:
        ticker = data['MKSC_SHRN_ISCD']
        ask_price_1 = data['ASKP1']
        bid_price_1 = data['BIDP1']
        accumulated_volume = data.get('ACML_VOL', 0)
        
        # 데이터 유효성 검사
        if ticker and ask_price_1 and bid_price_1 and accumulated_volume:
            # 최소 데이터가 모일 때까지 대기
            if accumulated_volume < 30:
                logging.info(f"{ticker}에 대한 데이터가 충분하지 않습니다. 더 많은 데이터를 기다립니다...")
                return

            real_time_predictions = real_time_prediction(stock_model, stock_scaler, ticker, time_step=30)
            if real_time_predictions is not None:
                logging.info(f"{ticker}에 대한 실시간 예측: {real_time_predictions}")
            else:
                logging.error(f"{ticker}에 대한 예측 실패.")
        else:
            logging.warning(f"{ticker}에 필요한 데이터가 누락되었습니다.")
    except KeyError as e:
        logging.error(f"WebSocket 데이터에서 누락된 키: {e}")

def real_time_prediction(model, scaler, ticker, time_step=60):
    """훈련된 모델을 사용하여 실시간 예측을 수행합니다."""
    real_time_data = get_real_time_data_with_retry(ticker)
    
    if real_time_data is None or len(real_time_data) < time_step:
        logging.warning(f"{ticker}에 대한 데이터가 충분하지 않습니다. 필요한 데이터: {time_step}, 받은 데이터: {len(real_time_data) if real_time_data is not None else 0}.")
        return None
    
    real_time_data = real_time_data[['stck_prpr']].rename(columns={'stck_prpr': 'close'})
    real_time_data, _ = normalize_data(real_time_data)
    scaled_real_time_data = real_time_data['close'].values.reshape(-1, 1)
    
    X_real_time, _ = create_dataset(scaled_real_time_data, time_step)
    if len(X_real_time) == 0:
        logging.warning(f"{ticker}에 대한 데이터셋 생성 후 충분한 데이터 포인트가 없습니다.")
        return None
    
    X_real_time = X_real_time.reshape(X_real_time.shape[0], X_real_time.shape[1], 1)
    
    predictions = model.predict(X_real_time)
    predictions = scaler.inverse_transform(predictions)
    return predictions

def run_trading(stock_codes, stock_model, stock_scaler):
    """
    Runs trading based on real-time predictions.

    Args:
    - stock_codes (list): List of stock codes to trade.
    - stock_model: Trained LSTM model.
    - stock_scaler: Scaler used for normalizing the data.
    """
    try:
        for stock_code in stock_codes:
            prediction = real_time_prediction(stock_model, stock_scaler, stock_code)

            if prediction is not None:
                # Execute trade based on the prediction
                execute_trade(stock_code, prediction)
            else:
                logging.error(f"Skipping trade execution for {stock_code} due to insufficient data or prediction error.")
    except Exception as e:
        logging.error(f"An error occurred during trade execution: {e}")

def run_websocket():
    """실시간 데이터를 수신하기 위해 WebSocket 클라이언트를 실행합니다."""
    websocket_url = "ws://ops.koreainvestment.com:21000"
    ws = websocket.WebSocketApp(websocket_url,
                                on_open=on_open,
                                on_message=on_message,
                                on_error=on_error,
                                on_close=on_close)
    ws.run_forever()

if __name__ == '__main__':
    try:
        stock_model, stock_scaler = load_lstm_model_and_scaler(stock_model_path, stock_scaler_path)
        stock_codes = get_stock_codes_from_csv(os.path.join(data_dir, 'stock_codes.csv'))

        run_trading(stock_codes, stock_model, stock_scaler)
        run_websocket()
    except Exception as e:
        logging.error(f"An error occurred in the main loop: {e}")
