import logging
import pandas as pd
from datetime import datetime
from helper_functions import get_available_cash, get_owned_quantity
from websocket_client import connect_to_websocket
import json

def generate_trading_signals(stock_data):
    # 이동 평균 계산
    stock_data['SMA_20'] = stock_data['Close'].rolling(window=20).mean()
    stock_data['SMA_50'] = stock_data['Close'].rolling(window=50).mean()

    # 신호 초기화
    stock_data['signal'] = 0
    
    # SMA_20과 SMA_50이 모두 NaN이 아닌 경우에만 신호 생성
    valid_indices = stock_data.index[(~stock_data['SMA_20'].isna()) & (~stock_data['SMA_50'].isna())]
    stock_data.loc[valid_indices, 'signal'] = (stock_data.loc[valid_indices, 'SMA_20'] > stock_data.loc[valid_indices, 'SMA_50']).astype(int)
    
    # 매수/매도 포지션 계산
    stock_data['position'] = stock_data['signal'].diff()
    
    return stock_data

def execute_trades(stock_data):
    try:
        # 매수/매도 신호 필터링
        buy_signals = stock_data[stock_data['position'] == 1]
        sell_signals = stock_data[stock_data['position'] == -1]

        # WebSocket 연결
        ws = connect_to_websocket()

        # 가용 자금 확인
        available_cash = get_available_cash()

        # 매수 신호 처리
        for index, signal in buy_signals.iterrows():
            stock_price = signal['Close']
            quantity = calculate_order_quantity(available_cash, stock_price)
            if quantity > 0:
                order = {
                    'type': 'buy',
                    'stock_code': signal.get('stock_code', 'UNKNOWN'),  # stock_code 확인
                    'price': stock_price,
                    'quantity': quantity
                }
                ws.send(json.dumps(order))
                available_cash -= stock_price * quantity

        # 매도 신호 처리
        for index, signal in sell_signals.iterrows():
            stock_price = signal['Close']
            owned_quantity = get_owned_quantity(signal.get('stock_code', 'UNKNOWN'))  # stock_code 확인
            if owned_quantity > 0:
                order = {
                    'type': 'sell',
                    'stock_code': signal.get('stock_code', 'UNKNOWN'),  # stock_code 확인
                    'price': stock_price,
                    'quantity': owned_quantity
                }
                ws.send(json.dumps(order))

    except Exception as e:
        logging.error(f"An error occurred during trade execution: {e}")

def calculate_order_quantity(available_cash, stock_price, risk_factor=0.01, max_portfolio_percent=0.05):
    try:
        max_investment = available_cash * risk_factor
        max_quantity = max_investment // stock_price

        max_portfolio_allocation = available_cash * max_portfolio_percent
        max_quantity = min(max_quantity, max_portfolio_allocation // stock_price)

        return max(1, int(max_quantity))
    except Exception as e:
        logging.error(f"An error occurred during order quantity calculation: {e}")
        return 0  # 오류 발생 시 주문 수량을 0으로 설정

